#pragma once

class AudioAppDemo   : public AudioAppComponent
{
public:
    AudioAppDemo()
    {
        setAudioChannels (0, 2);

        createGUI();
        setSize (800, 600);

        midiInputChanged();
    }

    ~AudioAppDemo() override
    {
        disableAllMidiOutputs();
        disableAllMidiInputs();
        shutdownAudio();
    }


    void disableAllMidiInputs()
    {
        auto midiInputList = juce::MidiInput::getAvailableDevices();

        // Ensure all MIDI inputs are disabled
        for (auto& x : midiInputList)
            if (deviceManager.isMidiInputDeviceEnabled(x.identifier))
                deviceManager.setMidiInputDeviceEnabled(x.identifier, false);

        // Remove previously existing callback
        deviceManager.removeMidiInputDeviceCallback("", &midiInputCollector);
    }


    void midiInputChanged()
    {
        disableAllMidiInputs();

        int itemId = midiInputSelection.getSelectedId();
        jassert(itemId != 0);        // Ensure our test has a valid comboBox item selected
        int listIndex = itemId - 1;

        auto midiInputList = juce::MidiInput::getAvailableDevices();

        // Enable the requested MIDI input and get data from all enabled MIDI inputs
        deviceManager.setMidiInputDeviceEnabled(midiInputList[listIndex].identifier, true);
        deviceManager.addMidiInputDeviceCallback("", &midiInputCollector);
    }


    void disableAllMidiOutputs()
    {
        if (midiOutputDevice == nullptr)
            return;

        midiOutputDevice->clearAllPendingMessages();
        midiOutputDevice->stopBackgroundThread();
        midiOutputDevice.reset();
    }


    void midiOutputChanged()
    {
        disableAllMidiOutputs();

        int itemId = midiOutputSelection.getSelectedId();
        jassert(itemId != 0);        // Ensure our test has a valid comboBox item selected
        int listIndex = itemId - 1;

        auto midiOutputList = juce::MidiOutput::getAvailableDevices();

        midiOutputDevice = juce::MidiOutput::openDevice(midiOutputList[listIndex].identifier);
        jassert(midiOutputDevice);

        midiOutputDevice->startBackgroundThread();
    }


    void paint (Graphics& g) override
    {
        g.fillAll(getLookAndFeel().findColour (ResizableWindow::backgroundColourId));
    }


    void createGUI()
    {
        //
        // MIDI Inputs
        //
        auto midiInputList = juce::MidiInput::getAvailableDevices();
        int index = 1;

        for (auto x : midiInputList)
            midiInputSelection.addItem(x.name, index++);

        midiInputSelection.onChange = [this] { midiInputChanged(); };
        midiInputSelection.setSelectedId(1);

        midiInputLabel.setText("MIDI Input", NotificationType::dontSendNotification);

        addAndMakeVisible(midiInputSelection);
        addAndMakeVisible(midiInputLabel);

        //
        // MIDI Output
        //
        auto midiOutputList = juce::MidiOutput::getAvailableDevices();
        index = 1;

        for (auto x : midiOutputList)
            midiOutputSelection.addItem(x.name, index++);

        midiOutputSelection.onChange = [this] { midiOutputChanged(); };
        midiOutputSelection.setSelectedId(1);

        midiOutputLabel.setText("MIDI Output", NotificationType::dontSendNotification);

        addAndMakeVisible(midiOutputSelection);
        addAndMakeVisible(midiOutputLabel);

        setSize(400, 200);
    }


    void resized() override
    {
        midiInputSelection.setBounds(10, 30, 300, 30);
        midiInputLabel.setBounds(10, 0, 300, 30);

        midiOutputSelection.setBounds(400, 30, 300, 30);
        midiOutputLabel.setBounds(400, 0, 300, 30);
    }


    void prepareToPlay(int samplesPerBlockExpected, double newSampleRate) override
    {
        sampleRate = newSampleRate;
        expectedSamplesPerBlock = samplesPerBlockExpected;

        midiInputCollector.reset(newSampleRate);
    }


    void getNextAudioBlock(const AudioSourceChannelInfo& bufferToFill) override
    {
        const double current_time = Time::getMillisecondCounterHiRes();

        // Clear audio data
        bufferToFill.clearActiveBufferRegion();

        // Get MIDI input data
        midiInputBuffer.clear();
        midiInputCollector.removeNextBlockOfMessages(midiInputBuffer, expectedSamplesPerBlock);

        // Create MIDI output data
        midiOutBuffer.clear();

        for (auto x : midiInputBuffer)
        {
            auto message = x.getMessage();

            // Forward only note on/off messages
            if (!message.isNoteOnOrOff())
                continue;

            midiOutBuffer.addEvent(message, 0);
        }

        if (midiOutputDevice)
            midiOutputDevice->sendBlockOfMessages(midiOutBuffer, current_time, sampleRate);
    }


    void releaseResources() override
    {
    }

private:
    double                      sampleRate = 0.0;
    int                         expectedSamplesPerBlock = 0;

    // MIDI Inputs
    juce::ComboBox              midiInputSelection;
    juce::Label                 midiInputLabel;

    juce::MidiMessageCollector  midiInputCollector;
    juce::MidiBuffer            midiInputBuffer;

    // MIDI Outputs
    juce::ComboBox              midiOutputSelection;
    juce::Label                 midiOutputLabel;

    std::unique_ptr<MidiOutput> midiOutputDevice;
    juce::MidiBuffer            midiOutBuffer;

    JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (AudioAppDemo)
};
